import json

import os

from os import path


class SkivvyConfig:
    def __init__(self, d):
        self.d = d

    def __getattr__(self, item):
        if item not in self.d:
            raise AttributeError("%s not found in %s" % (item, self.d))
        else:
            return self.d[item]

    def get(self, item, default=None):
        return self.d.get(item, default)

    def as_dict(self):
        return dict(self.d)


def read_config(config_name):
    cwd = os.getcwd()
    config_file = path.join(cwd, config_name)

    with open(config_file) as fp:
        config_dict = json.load(fp)
    return SkivvyConfig(config_dict)



from typing import Dict, Callable
import requests
from skivvy.util import dict_util
from dataclasses import dataclass
from typing import Mapping, Any, Optional
import json
from skivvy.util.str_util import tojsonstr

_supported_methods = {
    "get",
    "post",
    "put",
    "patch",
    "delete",
    "options",
    "head",
    "connect",
}
_session = None
_NO_BODY_STATUS = {204, 205, 304}


@dataclass(frozen=True, slots=True)
class HttpEnvelope:
    """Immutable container for an HTTP response with lazy/optional JSON parsing."""

    status_code: int
    headers: Mapping[str, str]
    text: str
    encoding: str
    url: str
    elapsed: float

    def has_body(self) -> bool:
        return len(self.text) > 0

    def content_type(self) -> str:
        return self.header("Content-Type", "")

    def json(self) -> Optional[Any]:
        """
        Returns parsed JSON, or None if:
        - body is empty
        - parsing fails (ValueError)
        """
        if not self.has_body():
            return None
        if len(self.text.strip()) == 0:
            return None

        try:
            return json.loads(self.text)
        except ValueError:
            return None

    def header(self, name: str, default: Optional[str] = None) -> Optional[str]:
        # case-insensitive lookup
        for k, v in self.headers.items():
            if k.lower() == name.lower():
                return v
        return default

    @staticmethod
    def from_requests(resp: requests.Response) -> "HttpEnvelope":
        return HttpEnvelope(
            status_code=getattr(resp, "status_code", 0),
            headers=getattr(resp, "headers", {}),
            text=getattr(resp, "text", ""),
            encoding=getattr(resp, "encoding", "utf-8"),
            url=getattr(resp, "url", ""),
            elapsed=getattr(resp, "elapsed", -1),
        )


def initialize_session(session=None):
    global _session
    _session = session or requests.Session()


def execute(request: dict[str, object]) -> HttpEnvelope:
    method, payload = prepare_request_data(request)
    r = do_request(method, **payload)
    return HttpEnvelope.from_requests(r)


def prepare_request_data(request_data: dict[str, object]) -> tuple[str, dict]:
    remap = {"form": "data", "upload": "files", "query": "params", "body": "json"}
    request_data = dict_util.remap_keys(request_data, remap)
    method = request_data.pop("method", "").lower()
    return method, request_data


def do_request(method, **payload: Dict[str, Any]) -> requests.Request:
    assert method, "missing method"
    assert method in _supported_methods, f"unsupported method: {method}"

    request_function: Callable = getattr(_session, method)
    assert callable(request_function), f"Session function {method} is not callable"
    return request_function(**payload)


initialize_session()


# TODO: This kind of functionality was support in the old http_util - add something similar but better for being able to just turn it on and generate a test case
# log.debug("\n")
# log.debug("--- REQUEST ---")
# log.debug("%s: %s" % (method.upper(), url))
# if upload_file:
#     log.debug( "file: %s" % upload_file)
# elif data:
#     log.debug(" body: %s" % data)
# log.debug(" headers: %s" % tojsonstr(headers))
# log.debug("----------------")
# log.debug("\n")


# it's not possible to both upload a file & provide json data
# if upload_file:
#     r = http_verb(url, files=upload_file, headers=headers)
# else:
#     r = http_verb(url, data=data, headers=headers)

# log.debug("\n")
# log.debug("--- RESPONSE ---")
# log.debug("status: %s" % r.status_code)
# log.debug("json: %s" % tojsonstr(as_json(r)))
# log.debug("headers: %s" % tojsonstr(dict(r.headers)))
# log.debug("----------------")
# log.debug("\n")



import json
import os
from typing import NamedTuple, Dict, Any, ChainMap, Mapping

from skivvy.util import file_util


class Option(NamedTuple):
    key: str
    default: Any
    help: str = ""

    def from_dict(self, config_dict: Dict[str, Any]):
        return config_dict.get(self.key, self.default)


class Settings:
    LOG_LEVEL = Option("log_level", "INFO", "Logging level")
    BASE_URL = Option("base_url", "", "Base URL for requests")
    URL = Option("url", None, "Request URL")
    BRACE_EXPANSION = Option(
        "brace_expansion", False, "Enable brace expansion (see README.md)"
    )
    BRACE_EXPANSION_WARNINGS = Option(
        "brace_expansion_warnings",
        True,
        "Log a warning when brace expansion fails to resolve a variable",
    )
    BRACE_EXPANSION_STRICT = Option(
        "brace_expansion_strict",
        False,
        "Raise an exception when brace expansion fails to resolve a variable",
    )
    VALIDATE_VARIABLE_NAMES = Option(
        "validate_variable_names",
        True,
        "Require variables to have typical syntax (start with letter, contain only alphanumerics and _-.,/\\ characters)",
    )
    AUTO_COERCE = Option(
        "auto_coerce",
        True,
        "Automatically coerce types in brace expansion, comparing values etc",
    )
    METHOD = Option("method", "get", "HTTP method")
    STATUS = Option("status", None, "Will only be checked if specified in the test")
    RESPONSE = Option("response", {}, "Expected response body")
    RESPONSE_HEADERS = Option("response_headers", None, "Expected response headers")
    HEADERS = Option("headers", None, "Request headers")
    BODY = Option("body", None, "JSON Request body")
    FORM = Option("form", None, "Form body")
    UPLOAD = Option("upload", None, "File upload configuration")
    QUERY = Option("query", None, "Query parameters")
    CONTENT_TYPE = Option("content_type", "application/json", "Request content type")
    WRITE_HEADERS = Option("write_headers", {}, "Headers to write to files")
    READ_HEADERS = Option("read_headers", {}, "Headers to read from files")
    MATCH_SUBSETS = Option(
        "match_subsets", False, "Allow subset matching in verification"
    )
    SKIP_EMPTY_OBJECTS = Option(
        "skip_empty_objects",
        False,
        "When subset matching, skip verification for empty objects",
    )
    SKIP_EMPTY_ARRAYS = Option(
        "skip_empty_arrays",
        False,
        "When subset matching, skip verification for empty arrays",
    )
    MATCH_EVERY_ENTRY = Option(
        "match_every_entry", False, "Require every actual array entry to match the expected template"
    )
    MATCH_FALSINESS = Option(
        "match_falsiness", True, "Match falsy values in verification"
    )
    COLORIZE = Option("colorize", True, "Enable colored output")
    FAIL_FAST = Option("fail_fast", False, "Stop on first failure")
    MATCHERS = Option("matchers", None, "Directory containing custom matcher files")
    MATCHER_OPTIONS = Option("matcher_options", {}, "Per-matcher configuration options")


def get_all_settings() -> list[Option]:
    return [
        option for _name, option in vars(Settings).items() if isinstance(option, Option)
    ]


def get_settings_by_key() -> dict[str, Option]:
    return {option.key: option for option in get_all_settings()}


def coerce_override_value(raw: str) -> object:
    value = raw.strip()
    if value == "":
        return value

    lowered = value.lower()
    if lowered == "true":
        return True
    if lowered == "false":
        return False
    if lowered in ("none", "null"):
        return None

    try:
        return json.loads(value)
    except json.JSONDecodeError:
        return value


def parse_cli_overrides(raw_overrides: list[str] | str | None) -> dict[str, object]:
    if raw_overrides is None:
        return {}

    values = raw_overrides if isinstance(raw_overrides, list) else [raw_overrides]
    known_settings = get_settings_by_key()
    overrides: dict[str, object] = {}
    for raw in values:
        key, sep, value = raw.partition("=")
        key = key.strip()
        if sep != "=" or key == "":
            raise ValueError(
                f'Invalid --set value "{raw}". Expected format: key=value'
            )
        if key not in known_settings:
            available = ", ".join(sorted(known_settings.keys()))
            raise ValueError(
                f'Unknown setting "{key}" passed via --set. Supported settings: {available}'
            )
        overrides[key] = coerce_override_value(value)

    return overrides


def env_var_name_for_option(option: Option, prefix: str = "SKIVVY_") -> str:
    normalized_key = "".join(ch if ch.isalnum() else "_" for ch in option.key.upper())
    return f"{prefix}{normalized_key}"


def parse_env_overrides(
    env: Mapping[str, str] | None = None, prefix: str = "SKIVVY_"
) -> dict[str, object]:
    source = os.environ if env is None else env
    overrides: dict[str, object] = {}
    for option in get_all_settings():
        env_key = env_var_name_for_option(option, prefix=prefix)
        if env_key in source:
            overrides[option.key] = coerce_override_value(source[env_key])
    return overrides


def create_test_config(*dicts: Dict[str, object]) -> Mapping[str, object]:
    """Creates a dict-like test configuration from multiple dictionaries
    Priority order:
    1. command-line arguments
    2. current test fields
    3. config file
    4. default values
    """
    defaults = {
        option.key: option.default for option in get_all_settings() if option.default
    }
    priority = [*dicts, defaults]
    d = ChainMap({}, *priority)
    return dict(**d)


def conf_get(d, option: Option):
    return d.get(option.key, option.default)


def create_testcase(*sources: Dict[str, object] | str) -> Mapping[str, object]:
    """Creates a testcase by merging any number of configurations into one single object.
    A source can either be a string or a dict, strings will be interpreted as paths to json files, the
    output is created by merging and potentially overriding fields by creating a chained map.
    """
    dicts: list[Dict[str, object]] = []
    for source in sources:
        if isinstance(source, str):
            source_dict = file_util.parse_json(source)
        elif isinstance(source, dict):
            source_dict = source
        elif hasattr(source, "as_dict") and callable(getattr(source, "as_dict")):
            source_dict = source.as_dict()
        else:
            source_dict = dict(source)

        dicts.append(source_dict)

    return create_test_config(*dicts)



import os

from skivvy.util import str_util, scope, file_util, log


def is_env_variable_name(variable_name: str) -> bool:
    return variable_name.lower().startswith("env.")


def brace_expand_string(s, **kwargs):
    if not isinstance(s, str):
        return s

    warn = kwargs.get("warn", True)
    strict = kwargs.get("strict", False)

    pattern = str_util.compile_regexp(brace_expansion_regexp)
    try:
        s = str_util.expand_string(s, pattern, resolve_funcs=resolvers)
    except ValueError as e:
        if warn:
            log.warning(f"[yellow]Brace expansion warning:[/yellow] {e}")
        if strict:
            raise
        return s

    auto_coerce_func = kwargs.get("auto_coerce_func", lambda v: v)
    return auto_coerce_func(s)


def variable_resolver(variable_name):
    if is_env_variable_name(variable_name):
        return None
    if scope.has(variable_name):
        return scope.fetch(variable_name)
    return None


def env_resolver(variable_name):
    prefix = "env."
    if not is_env_variable_name(variable_name):
        return None

    env_name = variable_name[len(prefix):]
    if env_name == "":
        return None
    return os.environ.get(env_name)


def file_resolver(variable_name):
    if is_env_variable_name(variable_name):
        return None
    if os.path.isfile(variable_name):
        return file_util.read_file_contents(variable_name)
    return None


resolvers = [env_resolver, variable_resolver, file_resolver]
brace_expansion_regexp = r"<([^<>]+)>"



import collections.abc
from functools import wraps
from operator import itemgetter


def subset(d: dict, keys: list[str], include_none=True) -> dict:
    """Returns a subset of a dictionary whose keys are included in keys."""

    def should_include(key):
        if include_none:
            return key in d
        else:
            return d.get(key, None) is not None

    return {k: d[k] for k in keys if should_include(k)}


def remap_keys(d: dict, remap: dict[str, str]) -> dict:
    """Returns a dictionary where keys are renamed according to remap (e.g. {'old: 'new'}).
    Other keys are left alone and returned in the new dictionary."""
    return {remap.get(k, k): v for k, v in d.items()}


def map_nested_dicts_py(d, func):
    if isinstance(d, collections.abc.Mapping):
        return {k: map_nested_dicts_py(v, func) for k, v in d.items()}
    else:
        return func(d)


def wrap_in_tuple(fn):
    @wraps(fn)
    def _inner(d, *keys):
        if len(keys) == 0:
            return ()
        items = fn(d, *keys)
        return items if isinstance(items, tuple) else (items,)

    return _inner


@wrap_in_tuple
def get_all(d, *keys):
    """
    retrieves all keys from a dict or raises an error
    """
    _get = itemgetter(*keys)
    return _get(d)


@wrap_in_tuple
def get_many(d, *keys):
    """
    returns as many keys as it finds from a dict
    """
    return filter_null_from_list([d.get(key) for key in keys])


def filter_null_from_list(l):
    """
    removes all values from a list that are None
    """
    return filter(lambda x: x is not None, l)


def filter_null_from_dict(d):
    """
    removes all values from a dict that are None
    """
    return {k: v for k, v in d.items() if v is not None}
